import enum
import typing

from modern_di import Container
from modern_di.providers import AbstractProvider


T_co = typing.TypeVar("T_co", covariant=True)
P = typing.ParamSpec("P")


class ContextAdapter(AbstractProvider[T_co]):
    __slots__ = [*AbstractProvider.BASE_SLOTS, "_function"]

    def __init__(
        self,
        scope: enum.IntEnum,
        function: typing.Callable[..., T_co],
    ) -> None:
        super().__init__(scope)
        self._function = function

    async def async_resolve(self, container: Container) -> T_co:
        return self._function(**container.find_container(self.scope).context)

    def sync_resolve(self, container: Container) -> T_co:
        return self._function(**container.find_container(self.scope).context)



import contextlib
import typing

import fastapi
import httpx
import pytest
from asgi_lifespan import LifespanManager

from modern_di_fastapi import setup_di


@contextlib.asynccontextmanager
async def lifespan(app_: fastapi.FastAPI) -> typing.AsyncIterator[None]:
    container = setup_di(app_)
    async with container:
        yield


@pytest.fixture
async def app() -> typing.AsyncIterator[fastapi.FastAPI]:
    app_ = fastapi.FastAPI(lifespan=lifespan)
    async with LifespanManager(app_):
        yield app_


@pytest.fixture
def client(app: fastapi.FastAPI) -> httpx.AsyncClient:
    return httpx.AsyncClient(transport=httpx.ASGITransport(app=app), base_url="http://test")



import contextlib
import enum
import inspect
import typing

from modern_di import Container
from modern_di.providers.abstract import AbstractCreatorProvider


T_co = typing.TypeVar("T_co", covariant=True)
P = typing.ParamSpec("P")


class Resource(AbstractCreatorProvider[T_co]):
    __slots__ = [*AbstractCreatorProvider.BASE_SLOTS, "_creator", "_args", "_kwargs", "_is_async"]

    def _is_creator_async(
        self,
        _: contextlib.AbstractContextManager[T_co] | contextlib.AbstractAsyncContextManager[T_co],
    ) -> typing.TypeGuard[contextlib.AbstractAsyncContextManager[T_co]]:
        return self._is_async

    def __init__(
        self,
        scope: enum.IntEnum,
        creator: typing.Callable[P, typing.Iterator[T_co] | typing.AsyncIterator[T_co]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        new_creator: typing.Any
        if inspect.isasyncgenfunction(creator):
            self._is_async = True
            new_creator = contextlib.asynccontextmanager(creator)
        elif inspect.isgeneratorfunction(creator):
            self._is_async = False
            new_creator = contextlib.contextmanager(creator)
        else:
            msg = "Unsupported resource type"
            raise RuntimeError(msg)

        super().__init__(scope, new_creator, *args, **kwargs)

    async def async_resolve(self, container: Container) -> T_co:
        container = container.find_container(self.scope)
        if (override := container.fetch_override(self.provider_id)) is not None:
            return typing.cast(T_co, override)

        provider_state = container.fetch_provider_state(
            self.provider_id, is_async_resource=self._is_async, is_lock_required=True
        )
        if provider_state.instance is not None:
            return typing.cast(T_co, provider_state.instance)

        assert provider_state.provider_lock
        await provider_state.provider_lock.acquire()

        try:
            if provider_state.instance is not None:
                return typing.cast(T_co, provider_state.instance)

            _intermediate_ = await self._async_build_creator(container)

            if self._is_creator_async(self._creator):  # type: ignore[arg-type]
                provider_state.context_stack = contextlib.AsyncExitStack()
                provider_state.instance = await provider_state.context_stack.enter_async_context(_intermediate_)
            else:
                provider_state.context_stack = contextlib.ExitStack()
                provider_state.instance = provider_state.context_stack.enter_context(_intermediate_)
        finally:
            provider_state.provider_lock.release()

        return typing.cast(T_co, provider_state.instance)

    def sync_resolve(self, container: Container) -> T_co:
        container = container.find_container(self.scope)
        if (override := container.fetch_override(self.provider_id)) is not None:
            return typing.cast(T_co, override)

        provider_state = container.fetch_provider_state(self.provider_id)
        if provider_state.instance is not None:
            return typing.cast(T_co, provider_state.instance)

        if self._is_async:
            msg = "Async resource cannot be resolved synchronously"
            raise RuntimeError(msg)

        _intermediate_ = self._sync_build_creator(container)

        provider_state.context_stack = contextlib.ExitStack()
        provider_state.instance = provider_state.context_stack.enter_context(
            typing.cast(contextlib.AbstractContextManager[typing.Any], _intermediate_)
        )

        return typing.cast(T_co, provider_state.instance)



import pytest

from modern_di import Container, Scope, providers
from tests.creators import create_async_resource, create_sync_resource


async_resource = providers.Resource(Scope.APP, create_async_resource)
sync_resource = providers.Resource(Scope.APP, create_sync_resource)
mapping = providers.Dict(Scope.APP, dep1=async_resource, dep2=sync_resource)


async def test_dict() -> None:
    async with Container(scope=Scope.APP, context={"option": "app"}) as app_container:
        mapping1 = await mapping.async_resolve(app_container)
        mapping2 = mapping.sync_resolve(app_container)
        resource1 = await async_resource.async_resolve(app_container)
        resource2 = sync_resource.sync_resolve(app_container)
        assert mapping1 == mapping2 == {"dep1": resource1, "dep2": resource2}


async def test_dict_wrong_scope() -> None:
    request_factory_ = providers.Factory(Scope.REQUEST, lambda: "")
    with pytest.raises(RuntimeError, match="Scope of dependency cannot be more than scope of dependent"):
        providers.Dict(Scope.APP, dep1=request_factory_)



import pytest

from modern_di import Container, Scope, providers
from tests.creators import create_async_resource, create_sync_resource


async_resource = providers.Resource(Scope.APP, create_async_resource)
sync_resource = providers.Resource(Scope.APP, create_sync_resource)
sequence = providers.List(Scope.APP, async_resource, sync_resource)


async def test_list() -> None:
    async with Container(scope=Scope.APP, context={"option": "app"}) as app_container:
        sequence1 = await sequence.async_resolve(app_container)
        sequence2 = sequence.sync_resolve(app_container)
        resource1 = await async_resource.async_resolve(app_container)
        resource2 = sync_resource.sync_resolve(app_container)
        assert sequence1 == sequence2 == [resource1, resource2]


async def test_list_wrong_scope() -> None:
    request_factory_ = providers.Factory(Scope.REQUEST, lambda: "")
    with pytest.raises(RuntimeError, match="Scope of dependency cannot be more than scope of dependent"):
        providers.List(Scope.APP, request_factory_)



import typing

import fastapi
import httpx
from modern_di import Scope, providers
from starlette import status

from modern_di_fastapi import Provide
from tests.dependencies import DependentCreator, SimpleCreator, context_adapter_function


app_factory = providers.Factory(Scope.APP, SimpleCreator, dep1="original")
request_factory = providers.Factory(Scope.REQUEST, DependentCreator, dep1=app_factory.cast)
context_adapter = providers.ContextAdapter(Scope.REQUEST, context_adapter_function)


async def test_read_main(client: httpx.AsyncClient, app: fastapi.FastAPI) -> None:
    @app.get("/")
    async def read_root(
        app_factory_instance: typing.Annotated[SimpleCreator, Provide(app_factory)],
        request_factory_instance: typing.Annotated[DependentCreator, Provide(request_factory)],
        method: typing.Annotated[str, Provide(context_adapter)],
    ) -> str:
        assert isinstance(app_factory_instance, SimpleCreator)
        assert isinstance(request_factory_instance, DependentCreator)
        return method

    response = await client.get("/")
    assert response.status_code == status.HTTP_200_OK
    assert response.json() == "GET"



import datetime

from modern_di import Container, Scope, providers


def context_adapter_function(*, now: datetime.datetime, **_: object) -> datetime.datetime:
    return now


context_adapter = providers.ContextAdapter(Scope.APP, context_adapter_function)
request_context_adapter = providers.ContextAdapter(Scope.REQUEST, context_adapter_function)


async def test_context_adapter() -> None:
    now = datetime.datetime.now(tz=datetime.timezone.utc)
    async with Container(scope=Scope.APP, context={"now": now}) as app_container:
        instance1 = await context_adapter.async_resolve(app_container)
        instance2 = context_adapter.sync_resolve(app_container)
        assert instance1 is instance2 is now


async def test_context_adapter_in_request_scope() -> None:
    now = datetime.datetime.now(tz=datetime.timezone.utc)
    async with (
        Container(scope=Scope.APP) as app_container,
        app_container.build_child_container(context={"now": now}) as request_container,
    ):
        instance1 = await request_context_adapter.async_resolve(request_container)
        instance2 = request_context_adapter.sync_resolve(request_container)
        assert instance1 is instance2 is now



import enum
import typing

from modern_di import Container
from modern_di.providers.abstract import AbstractProvider


T_co = typing.TypeVar("T_co", covariant=True)
P = typing.ParamSpec("P")


class Selector(AbstractProvider[T_co]):
    __slots__ = [*AbstractProvider.BASE_SLOTS, "_function", "_providers"]

    def __init__(
        self, scope: enum.IntEnum, function: typing.Callable[..., str], **providers: AbstractProvider[T_co]
    ) -> None:
        super().__init__(scope)
        self._check_providers_scope(providers.values())
        self._function: typing.Final = function
        self._providers: typing.Final = providers

    async def async_resolve(self, container: Container) -> T_co:
        container = container.find_container(self.scope)
        selected_key = self._function(**container.context)
        if selected_key not in self._providers:
            msg = f"No provider matches {selected_key}"
            raise RuntimeError(msg)

        return await self._providers[selected_key].async_resolve(container)

    def sync_resolve(self, container: Container) -> T_co:
        container = container.find_container(self.scope)
        selected_key = self._function(**container.context)
        if selected_key not in self._providers:
            msg = f"No provider matches {selected_key}"
            raise RuntimeError(msg)

        return self._providers[selected_key].sync_resolve(container)



import pytest

from modern_di import Container, Scope, providers


def selector_function(*, option: str, **_: object) -> "str":
    return option


app_factory = providers.Factory(Scope.APP, lambda: "app")
request_factory = providers.Factory(Scope.APP, lambda: "request")
app_selector = providers.Selector(Scope.APP, selector_function, app=app_factory, request=request_factory)
request_selector = providers.Selector(Scope.REQUEST, selector_function, app=app_factory, request=request_factory)


async def test_selector() -> None:
    async with Container(scope=Scope.APP, context={"option": "app"}) as app_container:
        instance1 = await app_selector.async_resolve(app_container)
        instance2 = app_selector.sync_resolve(app_container)
        assert instance1 == instance2 == "app"


async def test_selector_in_request_scope() -> None:
    async with (
        Container(scope=Scope.APP) as app_container,
        app_container.build_child_container(context={"option": "request"}) as request_container,
    ):
        instance1 = await request_selector.async_resolve(request_container)
        instance2 = request_selector.sync_resolve(request_container)
        assert instance1 == instance2 == "request"


async def test_selector_no_match() -> None:
    async with Container(scope=Scope.APP, context={"option": "wrong"}) as app_container:
        with pytest.raises(RuntimeError, match="No provider matches wrong"):
            await app_selector.async_resolve(app_container)

        with pytest.raises(RuntimeError, match="No provider matches wrong"):
            app_selector.sync_resolve(app_container)


async def test_selector_wrong_scope() -> None:
    request_factory_ = providers.Factory(Scope.REQUEST, lambda: "")
    with pytest.raises(RuntimeError, match="Scope of dependency cannot be more than scope of dependent"):
        providers.Selector(Scope.APP, lambda: "", request=request_factory_)



from modern_di.providers.abstract import AbstractProvider
from modern_di.providers.context_adapter import ContextAdapter
from modern_di.providers.dict import Dict
from modern_di.providers.factory import Factory
from modern_di.providers.list import List
from modern_di.providers.resource import Resource
from modern_di.providers.selector import Selector
from modern_di.providers.singleton import Singleton


__all__ = [
    "AbstractProvider",
    "ContextAdapter",
    "Factory",
    "Dict",
    "List",
    "Selector",
    "Singleton",
    "Resource",
]



import dataclasses
import enum
import typing

import fastapi
from modern_di import Container, Scope, providers


T_co = typing.TypeVar("T_co", covariant=True)


def setup_di(app: fastapi.FastAPI, scope: enum.IntEnum = Scope.APP) -> Container:
    app.state.di_container = Container(scope=scope)
    return app.state.di_container


def fetch_di_container(app: fastapi.FastAPI) -> Container:
    return typing.cast(Container, app.state.di_container)


async def build_request_container(request: fastapi.Request) -> typing.AsyncIterator[Container]:
    container: Container = fetch_di_container(request.app)
    async with container.build_child_container(context={"request": request}) as request_container:
        yield request_container


@dataclasses.dataclass(slots=True, frozen=True)
class Dependency(typing.Generic[T_co]):
    dependency: providers.AbstractProvider[T_co]

    async def __call__(
        self, request_container: typing.Annotated[Container, fastapi.Depends(build_request_container)]
    ) -> T_co:
        return await self.dependency.async_resolve(request_container)


def Provide(dependency: providers.AbstractProvider[T_co], *, use_cache: bool = True) -> T_co:  # noqa: N802
    return typing.cast(T_co, fastapi.Depends(dependency=Dependency(dependency), use_cache=use_cache))



import typing

from modern_di import Container
from modern_di.providers.abstract import AbstractCreatorProvider, AbstractProvider


if typing.TYPE_CHECKING:
    import typing_extensions


T = typing.TypeVar("T")
P = typing.ParamSpec("P")


class BaseGraph:
    providers: dict[str, AbstractProvider[typing.Any]]

    def __new__(cls, *_: typing.Any, **__: typing.Any) -> "typing_extensions.Self":  # noqa: ANN401
        msg = f"{cls.__name__} cannot not be instantiated"
        raise RuntimeError(msg)

    @classmethod
    def get_providers(cls) -> dict[str, AbstractProvider[typing.Any]]:
        if not hasattr(cls, "providers"):
            cls.providers = {k: v for k, v in cls.__dict__.items() if isinstance(v, AbstractProvider)}

        return cls.providers

    @classmethod
    async def async_resolve_creators(cls, container: Container) -> None:
        for provider in cls.get_providers().values():
            if isinstance(provider, AbstractCreatorProvider) and provider.scope == container.scope:
                await provider.async_resolve(container)

    @classmethod
    def sync_resolve_creators(cls, container: Container) -> None:
        for provider in cls.get_providers().values():
            if isinstance(provider, AbstractCreatorProvider) and provider.scope == container.scope:
                provider.sync_resolve(container)



import pytest

from modern_di import BaseGraph, Container, Scope, providers
from tests.creators import create_async_resource, create_sync_resource


async_resource = providers.Resource(Scope.APP, create_async_resource)
sync_resource = providers.Resource(Scope.APP, create_sync_resource)


async def test_graph_async_resolve_creators() -> None:
    class DIGraph(BaseGraph):
        async_resource = async_resource
        sync_resource = sync_resource

    async with Container(scope=Scope.APP) as app_container:
        await DIGraph.async_resolve_creators(app_container)

        assert len(app_container._provider_states) == 2  # noqa: SLF001, PLR2004
    assert len(app_container._provider_states) == 0  # noqa: SLF001


def test_graph_sync_resolve_creators() -> None:
    class DIGraph(BaseGraph):
        sync_resource = sync_resource

    with Container(scope=Scope.APP) as app_container:
        DIGraph.sync_resolve_creators(app_container)

        assert len(app_container._provider_states) == 1  # noqa: SLF001
    assert len(app_container._provider_states) == 0  # noqa: SLF001


def test_graph_cannot_be_instantiated() -> None:
    class DIGraph(BaseGraph):
        sync_resource = sync_resource

    with pytest.raises(RuntimeError, match="DIGraph cannot not be instantiated"):
        DIGraph()



import asyncio
import dataclasses
import typing

import pytest

from modern_di import Container, Scope, providers


@dataclasses.dataclass(kw_only=True, slots=True)
class SimpleCreator:
    dep1: str


@dataclasses.dataclass(kw_only=True, slots=True)
class DependentCreator:
    dep1: SimpleCreator


app_singleton = providers.Singleton(Scope.APP, SimpleCreator, dep1="original")
request_singleton = providers.Singleton(Scope.REQUEST, DependentCreator, dep1=app_singleton.cast)


async def test_app_singleton() -> None:
    async with Container(scope=Scope.APP) as app_container:
        singleton1 = await app_singleton.async_resolve(app_container)
        singleton2 = await app_singleton.async_resolve(app_container)
        assert singleton1 is singleton2

    with Container(scope=Scope.APP) as app_container:
        singleton3 = app_singleton.sync_resolve(app_container)
        singleton4 = app_singleton.sync_resolve(app_container)
        assert singleton3 is singleton4
        assert singleton3 is not singleton1

    async with Container(scope=Scope.APP) as app_container:
        singleton5 = await app_singleton.async_resolve(app_container)
        singleton6 = await app_singleton.async_resolve(app_container)
        assert singleton5 is singleton6
        assert singleton5 is not singleton3
        assert singleton5 is not singleton1


async def test_request_singleton() -> None:
    with Container(scope=Scope.APP) as app_container:
        with app_container.build_child_container() as request_container:
            instance1 = request_singleton.sync_resolve(request_container)
            instance2 = request_singleton.sync_resolve(request_container)
            assert instance1 is instance2

        async with app_container.build_child_container() as request_container:
            instance3 = await request_singleton.async_resolve(request_container)
            instance4 = await request_singleton.async_resolve(request_container)
            assert instance3 is instance4

        assert instance1 is not instance3


async def test_app_singleton_in_request_scope() -> None:
    with Container(scope=Scope.APP) as app_container:
        with app_container.build_child_container():
            singleton1 = await app_singleton.async_resolve(app_container)

        async with app_container.build_child_container():
            singleton2 = await app_singleton.async_resolve(app_container)

        assert singleton1 is singleton2


async def test_singleton_overridden() -> None:
    async with Container(scope=Scope.APP) as app_container:
        singleton1 = app_singleton.sync_resolve(app_container)

        app_singleton.override(SimpleCreator(dep1="override"), container=app_container)

        singleton2 = app_singleton.sync_resolve(app_container)
        singleton3 = await app_singleton.async_resolve(app_container)

        app_singleton.reset_override(app_container)

        singleton4 = app_singleton.sync_resolve(app_container)

        assert singleton2 is not singleton1
        assert singleton2 is singleton3
        assert singleton4 is singleton1


async def test_singleton_race_condition() -> None:
    calls: int = 0

    async def create_resource() -> typing.AsyncIterator[str]:
        nonlocal calls
        calls += 1
        await asyncio.sleep(0)
        yield ""

    resource = providers.Resource(Scope.APP, create_resource)
    factory_with_resource = providers.Singleton(Scope.APP, SimpleCreator, dep1=resource.cast)

    async def resolve_factory(container: Container) -> SimpleCreator:
        return await factory_with_resource.async_resolve(container)

    async with Container(scope=Scope.APP) as app_container:
        client1, client2 = await asyncio.gather(resolve_factory(app_container), resolve_factory(app_container))

    assert client1 is client2
    assert calls == 1


async def test_singleton_wrong_dependency_scope() -> None:
    def some_factory(_: SimpleCreator) -> None: ...

    request_singleton_ = providers.Singleton(Scope.REQUEST, SimpleCreator, dep1="original")
    with pytest.raises(RuntimeError, match="Scope of dependency cannot be more than scope of dependent"):
        providers.Singleton(Scope.APP, some_factory, request_singleton_.cast)



import enum
import typing

from modern_di import Container
from modern_di.providers.abstract import AbstractProvider


T_co = typing.TypeVar("T_co", covariant=True)


class Dict(AbstractProvider[dict[str, T_co]]):
    __slots__ = [*AbstractProvider.BASE_SLOTS, "_providers"]

    def __init__(self, scope: enum.IntEnum, **providers: AbstractProvider[T_co]) -> None:
        super().__init__(scope)
        self._check_providers_scope(providers.values())
        self._providers: typing.Final = providers

    async def async_resolve(self, container: Container) -> dict[str, T_co]:
        return {key: await provider.async_resolve(container) for key, provider in self._providers.items()}

    def sync_resolve(self, container: Container) -> dict[str, T_co]:
        return {key: provider.sync_resolve(container) for key, provider in self._providers.items()}



import enum
import typing

from modern_di import Container
from modern_di.providers.abstract import AbstractProvider


T_co = typing.TypeVar("T_co", covariant=True)


class List(AbstractProvider[list[T_co]]):
    __slots__ = [*AbstractProvider.BASE_SLOTS, "_providers"]

    def __init__(self, scope: enum.IntEnum, *providers: AbstractProvider[T_co]) -> None:
        super().__init__(scope)
        self._check_providers_scope(providers)
        self._providers: typing.Final = providers

    async def async_resolve(self, container: Container) -> list[T_co]:
        return [await x.async_resolve(container) for x in self._providers]

    def sync_resolve(self, container: Container) -> list[T_co]:
        return [x.sync_resolve(container) for x in self._providers]



import dataclasses

import fastapi


@dataclasses.dataclass(kw_only=True, slots=True)
class SimpleCreator:
    dep1: str


@dataclasses.dataclass(kw_only=True, slots=True)
class DependentCreator:
    dep1: SimpleCreator


def context_adapter_function(*, request: fastapi.Request, **_: object) -> str:
    return request.method



import datetime
import logging
import typing


logger = logging.getLogger(__name__)


async def create_async_resource() -> typing.AsyncIterator[datetime.datetime]:
    logger.debug("Async resource initiated")
    try:
        yield datetime.datetime.now(tz=datetime.timezone.utc)
    finally:
        logger.debug("Async resource destructed")


def create_sync_resource() -> typing.Iterator[datetime.datetime]:
    logger.debug("Resource initiated")
    try:
        yield datetime.datetime.now(tz=datetime.timezone.utc)
    finally:
        logger.debug("Resource destructed")



import abc
import enum
import itertools
import typing
import uuid

from modern_di import Container


T_co = typing.TypeVar("T_co", covariant=True)
R = typing.TypeVar("R")
P = typing.ParamSpec("P")


class AbstractProvider(typing.Generic[T_co], abc.ABC):
    BASE_SLOTS: typing.ClassVar = ["scope", "provider_id"]

    def __init__(self, scope: enum.IntEnum) -> None:
        self.scope = scope
        self.provider_id: typing.Final = str(uuid.uuid4())

    @abc.abstractmethod
    async def async_resolve(self, container: Container) -> T_co:
        """Resolve dependency asynchronously."""

    @abc.abstractmethod
    def sync_resolve(self, container: Container) -> T_co:
        """Resolve dependency synchronously."""

    @property
    def cast(self) -> T_co:
        return typing.cast(T_co, self)

    def _check_providers_scope(self, providers: typing.Iterable[typing.Any]) -> None:
        if any(x.scope > self.scope for x in providers if isinstance(x, AbstractProvider)):
            msg = "Scope of dependency cannot be more than scope of dependent"
            raise RuntimeError(msg)


class AbstractOverrideProvider(AbstractProvider[T_co], abc.ABC):
    def override(self, override_object: object, container: Container) -> None:
        container.override(self.provider_id, override_object)

    def reset_override(self, container: Container) -> None:
        container.reset_override(self.provider_id)


class AbstractCreatorProvider(AbstractOverrideProvider[T_co], abc.ABC):
    BASE_SLOTS: typing.ClassVar = [*AbstractProvider.BASE_SLOTS, "_args", "_kwargs", "_creator"]

    def __init__(
        self,
        scope: enum.IntEnum,
        creator: typing.Callable[P, typing.Any],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        super().__init__(scope)
        self._check_providers_scope(itertools.chain(args, kwargs.values()))
        self._creator: typing.Final = creator
        self._args: typing.Final = args
        self._kwargs: typing.Final = kwargs

    def _sync_build_creator(self, container: Container) -> typing.Any:  # noqa: ANN401
        return self._creator(
            *typing.cast(
                P.args, [x.sync_resolve(container) if isinstance(x, AbstractProvider) else x for x in self._args]
            ),
            **typing.cast(
                P.kwargs,
                {
                    k: v.sync_resolve(container) if isinstance(v, AbstractProvider) else v
                    for k, v in self._kwargs.items()
                },
            ),
        )

    async def _async_build_creator(self, container: Container) -> typing.Any:  # noqa: ANN401
        return self._creator(
            *typing.cast(
                P.args,
                [await x.async_resolve(container) if isinstance(x, AbstractProvider) else x for x in self._args],
            ),
            **typing.cast(
                P.kwargs,
                {
                    k: await v.async_resolve(container) if isinstance(v, AbstractProvider) else v
                    for k, v in self._kwargs.items()
                },
            ),
        )



import dataclasses

import pytest

from modern_di import Container, Scope, providers


@dataclasses.dataclass(kw_only=True, slots=True)
class SimpleCreator:
    dep1: str


@dataclasses.dataclass(kw_only=True, slots=True)
class DependentCreator:
    dep1: SimpleCreator


app_factory = providers.Factory(Scope.APP, SimpleCreator, dep1="original")
request_factory = providers.Factory(Scope.REQUEST, DependentCreator, dep1=app_factory.cast)


async def test_app_factory() -> None:
    async with Container(scope=Scope.APP) as app_container:
        instance1 = await app_factory.async_resolve(app_container)
        instance2 = await app_factory.async_resolve(app_container)
        assert instance1 is not instance2

    with Container(scope=Scope.APP) as app_container:
        instance3 = app_factory.sync_resolve(app_container)
        instance4 = app_factory.sync_resolve(app_container)
        assert instance3 is not instance4
        assert instance1 is not instance3


async def test_request_factory() -> None:
    with Container(scope=Scope.APP) as app_container:
        with app_container.build_child_container() as request_container:
            instance1 = request_factory.sync_resolve(request_container)
            instance2 = request_factory.sync_resolve(request_container)
            assert instance1 is not instance2

        async with app_container.build_child_container() as request_container:
            instance3 = await request_factory.async_resolve(request_container)
            instance4 = await request_factory.async_resolve(request_container)
            assert instance3 is not instance4

        assert instance1 is not instance3


async def test_app_factory_in_request_scope() -> None:
    with Container(scope=Scope.APP) as app_container:
        with app_container.build_child_container():
            instance1 = await app_factory.async_resolve(app_container)

        async with app_container.build_child_container():
            instance2 = await app_factory.async_resolve(app_container)

        assert instance1 is not instance2


async def test_factory_overridden() -> None:
    async with Container(scope=Scope.APP) as app_container:
        instance1 = app_factory.sync_resolve(app_container)

        app_factory.override(SimpleCreator(dep1="override"), container=app_container)

        instance2 = app_factory.sync_resolve(app_container)
        instance3 = await app_factory.async_resolve(app_container)
        assert instance1 is not instance2
        assert instance2 is instance3
        assert instance2.dep1 != instance1.dep1

        app_factory.reset_override(app_container)

        instance4 = app_factory.sync_resolve(app_container)

        assert instance4.dep1 == instance1.dep1

        assert instance3 is not instance4


async def test_factory_wrong_dependency_scope() -> None:
    def some_factory(_: SimpleCreator) -> None: ...

    request_factory_ = providers.Factory(Scope.REQUEST, SimpleCreator, dep1="original")
    with pytest.raises(RuntimeError, match="Scope of dependency cannot be more than scope of dependent"):
        providers.Singleton(Scope.APP, some_factory, request_factory_.cast)


async def test_factory_scope_is_not_initialized() -> None:
    async with Container(scope=Scope.APP) as app_container:
        with pytest.raises(RuntimeError, match="Scope REQUEST is not initialize"):
            await request_factory.async_resolve(app_container)



import enum

import pytest

from modern_di import Container, Scope


def test_container_wrong_init() -> None:
    with pytest.raises(RuntimeError, match="Only first scope can be used without parent_container"):
        Container(scope=Scope.REQUEST)


def test_container_not_opened() -> None:
    container = Container(scope=Scope.APP)
    with pytest.raises(RuntimeError, match="Enter the context first"):
        container.fetch_provider_state("some_id")


async def test_container_build_child_async() -> None:
    async with Container(scope=Scope.APP) as app_container, app_container.build_child_container() as request_container:
        assert request_container.scope == Scope.REQUEST
        assert app_container.scope == Scope.APP


def test_container_build_child_sync() -> None:
    with Container(scope=Scope.APP) as app_container, app_container.build_child_container() as request_container:
        assert request_container.scope == Scope.REQUEST
        assert app_container.scope == Scope.APP


def test_container_scope_limit_reached() -> None:
    class CustomScope(enum.IntEnum):
        APP = 1
        REQUEST = 2

    with Container(scope=CustomScope.APP) as app_container, app_container.build_child_container() as request_container:
        assert request_container.scope == CustomScope.REQUEST
        assert app_container.scope == CustomScope.APP

        with pytest.raises(RuntimeError, match="Max scope is reached, REQUEST"):
            request_container.build_child_container()



from modern_di_fastapi.main import Provide, build_request_container, fetch_di_container, setup_di


__all__ = [
    "Provide",
    "build_request_container",
    "fetch_di_container",
    "setup_di",
]



import contextlib
import enum
import types
import typing

from modern_di.provider_state import ProviderState


if typing.TYPE_CHECKING:
    import typing_extensions


T_co = typing.TypeVar("T_co", covariant=True)


class Container(contextlib.AbstractAsyncContextManager["Container"]):
    __slots__ = "scope", "parent_container", "context", "_is_async", "_provider_states", "_overrides"

    def __init__(
        self,
        *,
        scope: enum.IntEnum,
        parent_container: typing.Optional["Container"] = None,
        context: dict[str, typing.Any] | None = None,
    ) -> None:
        if scope.value != 1 and parent_container is None:
            msg = "Only first scope can be used without parent_container"
            raise RuntimeError(msg)

        self.scope = scope
        self.parent_container = parent_container
        self.context: dict[str, typing.Any] = context or {}
        self._is_async: bool | None = None
        self._provider_states: dict[str, ProviderState[typing.Any]] = {}
        self._overrides: dict[str, typing.Any] = {}

    def _exit(self) -> None:
        self._is_async = None
        self._provider_states = {}
        self._overrides = {}
        self.context = {}

    def _check_entered(self) -> None:
        if self._is_async is None:
            msg = "Enter the context first"
            raise RuntimeError(msg)

    def build_child_container(self, context: dict[str, typing.Any] | None = None) -> "typing_extensions.Self":
        self._check_entered()

        try:
            new_scope = self.scope.__class__(self.scope.value + 1)
        except ValueError as exc:
            msg = f"Max scope is reached, {self.scope.name}"
            raise RuntimeError(msg) from exc

        return self.__class__(scope=new_scope, parent_container=self, context=context)

    def find_container(self, scope: enum.IntEnum) -> "typing_extensions.Self":
        container = self
        if container.scope < scope:
            msg = f"Scope {scope.name} is not initialized"
            raise RuntimeError(msg)

        while container.scope > scope and container.parent_container:
            container = typing.cast("typing_extensions.Self", container.parent_container)
        return container

    def fetch_provider_state(
        self, provider_id: str, is_async_resource: bool = False, is_lock_required: bool = False
    ) -> ProviderState[typing.Any]:
        self._check_entered()
        if is_async_resource and self._is_async is False:
            msg = "Resolving async resource in sync container is not allowed"
            raise RuntimeError(msg)

        if provider_id not in self._provider_states:
            self._provider_states[provider_id] = ProviderState(is_lock_required=is_lock_required)

        return self._provider_states[provider_id]

    def override(self, provider_id: str, override_object: object) -> None:
        self._overrides[provider_id] = override_object

    def fetch_override(self, provider_id: str) -> object | None:
        return self._overrides.get(provider_id)

    def reset_override(self, provider_id: str | None = None) -> None:
        if provider_id is None:
            self._overrides = {}
        else:
            self._overrides.pop(provider_id, None)

    async def __aenter__(self) -> "Container":
        self._is_async = True
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        self._check_entered()
        for provider_state in reversed(self._provider_states.values()):
            await provider_state.async_tear_down()
        self._exit()

    def __enter__(self) -> "Container":
        self._is_async = False
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        self._check_entered()
        for provider_state in reversed(self._provider_states.values()):
            provider_state.sync_tear_down()
        self._exit()



import asyncio
import typing

import pytest

from modern_di import Container, Scope, providers
from tests.creators import create_async_resource, create_sync_resource


async_resource = providers.Resource(Scope.APP, create_async_resource)
sync_resource = providers.Resource(Scope.APP, create_sync_resource)


async def test_async_resource() -> None:
    async with Container(scope=Scope.APP) as app_container:
        async_resource1 = await async_resource.async_resolve(app_container)
        async_resource2 = await async_resource.async_resolve(app_container)
        assert async_resource1 is async_resource2

    async with Container(scope=Scope.APP) as app_container:
        async_resource3 = await async_resource.async_resolve(app_container)
        async_resource4 = async_resource.sync_resolve(app_container)
        assert async_resource3 is async_resource4
        assert async_resource3 is not async_resource1


async def test_async_resource_in_sync_container() -> None:
    with (
        Container(scope=Scope.APP) as app_container,
        pytest.raises(RuntimeError, match="Resolving async resource in sync container is not allowed"),
    ):
        await async_resource.async_resolve(app_container)


async def test_async_resource_calling_sync_exit() -> None:
    async with Container(scope=Scope.APP) as app_container:
        await async_resource.async_resolve(app_container)
        with pytest.raises(RuntimeError, match="Cannot tear down async context in `sync_tear_down`"):
            app_container.__exit__(None, None, None)


async def test_sync_resource() -> None:
    async with Container(scope=Scope.APP) as app_container:
        sync_resource1 = await sync_resource.async_resolve(app_container)
        sync_resource2 = await sync_resource.async_resolve(app_container)
        assert sync_resource1 is sync_resource2

    with Container(scope=Scope.APP) as app_container:
        sync_resource3 = sync_resource.sync_resolve(app_container)
        sync_resource4 = sync_resource.sync_resolve(app_container)
        assert sync_resource3 is sync_resource4
        assert sync_resource3 is not sync_resource1


async def test_async_resource_overridden() -> None:
    async with Container(scope=Scope.APP) as app_container:
        async_resource1 = await async_resource.async_resolve(app_container)

        async_resource.override("override", container=app_container)

        async_resource2 = async_resource.sync_resolve(app_container)
        async_resource3 = await async_resource.async_resolve(app_container)

        app_container.reset_override()

        async_resource4 = async_resource.sync_resolve(app_container)

        assert async_resource2 is not async_resource1
        assert async_resource2 is async_resource3
        assert async_resource4 is async_resource1


async def test_sync_resource_overridden() -> None:
    async with Container(scope=Scope.APP) as app_container:
        sync_resource1 = await sync_resource.async_resolve(app_container)

        sync_resource.override("override", container=app_container)

        sync_resource2 = sync_resource.sync_resolve(app_container)
        sync_resource3 = await sync_resource.async_resolve(app_container)

        app_container.reset_override()

        sync_resource4 = sync_resource.sync_resolve(app_container)

        assert sync_resource2 is not sync_resource1
        assert sync_resource2 is sync_resource3
        assert sync_resource4 is sync_resource1


async def test_async_resource_race_condition() -> None:
    calls: int = 0

    async def create_resource() -> typing.AsyncIterator[str]:
        nonlocal calls
        calls += 1
        await asyncio.sleep(0)
        yield ""

    resource = providers.Resource(Scope.APP, create_resource)

    async def resolve_resource(container: Container) -> str:
        return await resource.async_resolve(container)

    async with Container(scope=Scope.APP) as app_container:
        await asyncio.gather(resolve_resource(app_container), resolve_resource(app_container))

    assert calls == 1


async def test_resource_unsupported_creator() -> None:
    with pytest.raises(RuntimeError, match="Unsupported resource type"):
        providers.Resource(Scope.APP, None)  # type: ignore[arg-type]


async def test_async_resource_sync_resolve() -> None:
    async with Container(scope=Scope.APP) as app_container:
        with pytest.raises(RuntimeError, match="Async resource cannot be resolved synchronously"):
            async_resource.sync_resolve(app_container)



import asyncio
import contextlib
import typing


T_co = typing.TypeVar("T_co", covariant=True)


class ProviderState(typing.Generic[T_co]):
    __slots__ = "context_stack", "instance", "provider_lock"

    def __init__(self, is_lock_required: bool) -> None:
        self.context_stack: contextlib.AsyncExitStack | contextlib.ExitStack | None = None
        self.instance: T_co | None = None
        self.provider_lock: typing.Final = asyncio.Lock() if is_lock_required else None

    async def async_tear_down(self) -> None:
        if self.context_stack is None:
            return

        if isinstance(self.context_stack, contextlib.AsyncExitStack):
            await self.context_stack.aclose()
        else:
            self.context_stack.close()
        self.context_stack = None
        self.instance = None

    def sync_tear_down(self) -> None:
        if self.context_stack is None:
            return

        if isinstance(self.context_stack, contextlib.AsyncExitStack):
            msg = "Cannot tear down async context in `sync_tear_down`"
            raise RuntimeError(msg)  # noqa: TRY004

        self.context_stack.close()
        self.context_stack = None
        self.instance = None



import enum
import typing

from modern_di import Container
from modern_di.providers.abstract import AbstractCreatorProvider


T_co = typing.TypeVar("T_co", covariant=True)
P = typing.ParamSpec("P")


class Singleton(AbstractCreatorProvider[T_co]):
    __slots__ = [*AbstractCreatorProvider.BASE_SLOTS, "_creator"]

    def __init__(
        self,
        scope: enum.IntEnum,
        creator: typing.Callable[P, T_co],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        super().__init__(scope, creator, *args, **kwargs)

    async def async_resolve(self, container: Container) -> T_co:
        container = container.find_container(self.scope)
        if (override := container.fetch_override(self.provider_id)) is not None:
            return typing.cast(T_co, override)

        provider_state = container.fetch_provider_state(self.provider_id, is_lock_required=True)
        if provider_state.instance is not None:
            return typing.cast(T_co, provider_state.instance)

        assert provider_state.provider_lock
        await provider_state.provider_lock.acquire()

        try:
            if provider_state.instance is not None:
                return typing.cast(T_co, provider_state.instance)

            provider_state.instance = typing.cast(T_co, await self._async_build_creator(container))
        finally:
            provider_state.provider_lock.release()

        return provider_state.instance

    def sync_resolve(self, container: Container) -> T_co:
        container = container.find_container(self.scope)
        if (override := container.fetch_override(self.provider_id)) is not None:
            return typing.cast(T_co, override)

        provider_state = container.fetch_provider_state(self.provider_id)
        if provider_state.instance is not None:
            return typing.cast(T_co, provider_state.instance)

        provider_state.instance = self._sync_build_creator(container)
        return typing.cast(T_co, provider_state.instance)
